@@ -1,36 +1,32 @@
-#' @importFrom rlang list2 names2
+#' @importFrom rlang list2 names2 %||%
 NULL
 
 #' Bundling Sass layers
 #'
-#' Sass layers are a way to group a set of related Sass variable definitions,
-#' function/mixin declarations, and CSS rules into a single object. Use
-#' `sass_layer()` to create these objects, and `sass_bundle()` to combine
-#' two or more layers or bundles objects into a Sass bundle; this ability to be merged is
-#' the main benefit of using Sass layers versus lower-level forms of sass input.
-#' At a later time, Sass layers may be removed from Sass bundles
-#' by referencing the same name that was used when creating the Sass bundle.
+#' Sass layers provide a way to package Sass variables, rules, functions, and
+#' mixins in a structured and composable way that follows best Sass practices.
+#' Most importantly, when multiple `sass_layer()` are combined into a
+#' `sass_bundle()`, variable `defaults` for later layers are placed _before_
+#' earlier layers, effectively 'new' defaults through all the 'old' defaults.
 #'
-#' @md
-#' @param ... A collection of `sass_layer()`s and/or objects that [sass::as_sass()]
-#'   understands. Arguments should be provided in reverse priority order:
-#'   defaults, declarations, and rules in later layers will take precedence over
-#'   those of previous layers. Non-layer values will be converted to layers by
-#'   calling `sass_layer(rules = ...)`.
-#' @param defaults A suitable [sass::as_sass()] `input`. Intended for declaring
-#'   variables with `!default`. When layers are combined, defaults are merged in
-#'   reverse order; that is, `sass_bundle(layer1, layer2)` will include
+#' @param functions [sass::as_sass()] `input` intended for [Sass
+#'   functions](https://rstudio.github.io/sass/articles/sass.html#functions-1).
+#'   Functions are placed before `defaults` so that variable definitions may make
+#'   use of functions.
+#' @param defaults [sass::as_sass()] `input` intended for [variable
+#'   defaults](https://rstudio.github.io/sass/articles/sass.html#variables-1).
+#'   These variable defaults after placed after `functions` but before `mixins`.
+#'   When multiple layers are combined in a `sass_bundle()`, defaults are merged
+#'   in reverse order; that is, `sass_bundle(layer1, layer2)` will include
 #'   `layer2$defaults` before `layer1$defaults`.
-#' @param declarations A suitable [sass::as_sass()] `input`.  Intended for
-#'   function and mixin declarations, and variable declarations without
-#'   `!default`; not intended for actual CSS rules. These will be merged in
-#'   forward order; that is, `sass_bundle(layer1, layer2)` will include
-#'   `layer1$declarations` before `layer2$declarations`.
-#' @param rules A suitable [sass::as_sass()] `input`. Intended for actual CSS
-#'   rules. These will be merged in forward order; that is,
-#'   `sass_bundle(layer1, layer2)` will include `layer1$rules` before
-#'   `layer2$rules`.
-#' @param html_deps An HTML dependency (or a list of them).
+#' @param mixins [sass::as_sass()] `input` intended for [Sass
+#'   mixins](https://rstudio.github.io/sass/articles/sass.html#mixins-1). Mixins
+#'   are placed after `defaults`, but before `rules`.
+#' @param rules [sass::as_sass()] `input` intended for [Sass
+#'   rules](https://sass-lang.com/documentation/style-rules). Rules are placed last
+#'   (i.e., after `functions`, `defaults`, and `mixins`).
+#' @param html_deps An HTML dependency (or a list of them). This dependency
+#'   gets attached to the return value of [sass()]/[as_sass()].
 #' @param file_attachments A named character vector, representing file assets
 #'   that are referenced (using relative paths) from the sass in this layer. The
 #'   vector names should be a relative path, and the corresponding vector values
@@ -40,15 +36,17 @@ NULL
 #'   into the destination directory; the directory itself will not be copied.)
 #'   You can also omit the name, in which case that file or directory will be
 #'   copied directly into the output directory.
+#' @param declarations Deprecated, use `functions` or `mixins` instead.
 #' @param tags Deprecated. Preserve meta information using a key in `sass_bundle(KEY = val)`.
 #'   preserve simple metadata as layers are merged.
+#' @md
 #' @examples
 #' blue <- list(color = "blue !default")
 #' red <- list(color = "red !default")
 #' green <- list(color = "green !default")
 #'
 #' # a sass_layer() by itself is not very useful, it just defines some
-#' # SASS to place before (defaults) and after (declarations, rules)
+#' # SASS to place before (defaults) and after (rules)
 #' core <- sass_layer(defaults = blue, rules = "body { color: $color; }")
 #' core
 #' sass(core)
@@ -90,14 +88,22 @@ NULL
 #' @describeIn sass_layer Compose the parts of a single Sass layer. Object returned is a `sass_bundle()` with a single Sass layer
 #' @export
 sass_layer <- function(
+  functions = NULL,
   defaults = NULL,
-  declarations = NULL,
+  mixins = NULL,
   rules = NULL,
   html_deps = NULL,
   file_attachments = character(0),
+  declarations = NULL,
   tags = NULL
 ) {
 
+  # In a future release cycle (once bslib has been updated to not use
+  # declarations), then consider throwing this deprecation message
+  #if (!is.null(declarations)) {
+  #  .Deprecated(msg="`declarations` is deprecated. Please use `functions` or `mixins instead.`")
+  #}
+
   if (!missing(tags)) {
     .Deprecated(msg="`sass_layer(tags)` is deprecated. Please use a named layer in a `sass_bundle(NAME = layer)`")
   }
@@ -105,49 +111,83 @@ sass_layer <- function(
   # return a size 1 sass_bundle()
   as_sass_bundle(
     sass_layer_struct(
+      functions = functions,
       defaults = defaults,
-      declarations = declarations,
+      mixins = mixins,
       rules = rules,
+      declarations = declarations,
       html_deps = html_deps,
       file_attachments = file_attachments
     )
   )
 }
 
+
+#' @export
+#' @param file file path to a `.scss` file.
+#' @describeIn sass_layer Read in a `.scss` file with parse special `/*-- scss:(functions|defaults|rules|mixins) --*/` comments as relevant sections of a `sass_layer()`.
+sass_layer_file <- function(file) {
+  src <- readLines(file)
+  # https://github.com/quarto-dev/quarto-cli/blob/3d6063/src/command/render/sass.ts#L119-L123
+  pattern_key <- "scss:(functions|defaults|rules|mixins)"
+  pattern <- paste0("^/\\*--[ \\t]*", pattern_key, "[ \\t]*--\\*/$")
+  idx <- grep(pattern, src)
+  if (length(idx) == 0) {
+    return(sass_layer(rules = src))
+  }
+  types <- extract_group(src[idx], pattern_key)
+  utypes <- unique(types)
+  args <- rlang::set_names(vector("list", length(utypes)), utypes)
+  for (i in seq_along(idx)) {
+    start <- idx[i] + 1
+    end <- if (i == length(idx)) length(src) else (idx[i + 1] - 1)
+    type <- types[i]
+    args[[type]] <- c(args[[type]], src[seq.int(start, end)])
+  }
+  do.call(sass_layer, args)
+}
+
 #' Helps avoid sass_layer / sass_bundle inf recursion
 #' @return object of class `sass_layer`
 #' @noRd
 sass_layer_struct <- function(
+  functions = NULL,
   defaults = NULL,
-  declarations = NULL,
+  mixins = NULL,
   rules = NULL,
+  declarations = NULL,
   html_deps = NULL,
-  file_attachments = character(0)
+  file_attachments = character(0),
+  validate = TRUE
 ) {
 
-  validate_layer_param(defaults, "defaults")
-  validate_layer_param(declarations, "declarations")
-  validate_layer_param(rules, "rules")
+  if (validate) {
+    validate_layer_param(functions, "functions")
+    validate_layer_param(defaults, "defaults")
+    validate_layer_param(mixins, "mixins")
+    validate_layer_param(rules, "rules")
+    validate_layer_param(declarations, "declarations")
 
-  validate_attachments(file_attachments)
-
-  if (!is.null(html_deps)) {
-    if (is_dependency_maybe(html_deps)) {
-      html_deps <- list(html_deps)
-    }
-    if (!is.list(html_deps)) {
-      stop("`html_deps` must be a collection of htmlDependency() and/or tagFunction() objects")
-    }
-    is_dependency <- vapply(html_deps, is_dependency_maybe, logical(1))
-    if (any(!is_dependency)) {
-      stop("`html_deps` must be a collection of htmlDependency() and/or tagFunction() objects")
+    if (!is.null(html_deps)) {
+      if (is_dependency_maybe(html_deps)) {
+        html_deps <- list(html_deps)
+      }
+      if (!is.list(html_deps)) {
+        stop("`html_deps` must be a collection of htmlDependency() and/or tagFunction() objects")
+      }
+      is_dependency <- vapply(html_deps, is_dependency_maybe, logical(1))
+      if (any(!is_dependency)) {
+        stop("`html_deps` must be a collection of htmlDependency() and/or tagFunction() objects")
+      }
     }
   }
 
   layer <- list(
+    functions = functions,
     defaults = defaults,
-    declarations = declarations,
+    mixins = mixins,
     rules = rules,
+    declarations = declarations,
     html_deps = html_deps,
     file_attachments = file_attachments
   )
@@ -246,6 +286,11 @@ sass_bundle_struct <- function(layers = list()) {
 }
 
 #' @describeIn sass_layer Collect `sass_bundle()` and/or `sass_layer()` objects. Unnamed Sass bundles will be concatenated together, preserving their internal name structures. Named Sass bundles will be condensed into a single Sass layer for easier removal from the returned Sass bundle.
+#' @param ... A collection of `sass_layer()`s and/or objects that [sass::as_sass()]
+#'   understands. Arguments should be provided in reverse priority order:
+#'   defaults, declarations, and rules in later layers will take precedence over
+#'   those of previous layers. Non-layer values will be converted to layers by
+#'   calling `sass_layer(rules = ...)`.
 #' @export
 sass_bundle <- function(...) {
   layers <- dropNulls(list2(...))
@@ -298,15 +343,11 @@ sass_bundle_remove <- function(bundle, name) {
 }
 
 
-
-
 # sass_layer Check if `x` is a Sass layer object
 is_sass_layer <- function(x) {
   inherits(x, "sass_layer")
 }
 
-
-
 #' @describeIn sass_layer Check if `x` is a Sass bundle object
 #' @param x object to inspect
 #' @export
@@ -315,7 +356,6 @@ is_sass_bundle <- function(x) {
 }
 
 
-
 #' Sass Bundle to Single Sass Layer
 #'
 #' Converts a [sass_bundle()] to a single Sass layer object.
@@ -329,54 +369,37 @@ is_sass_bundle <- function(x) {
 as_sass_layer <- function(x) {
   if (is_sass_layer(x)) return(x)
   # sass_bundle(x) will auto upgrade to a sass bundle object
-  Reduce(function(y1, y2) { sass_layers_join(y1, y2) }, sass_bundle(x)$layers)
-}
-sass_layers_join <- function(layer1, layer2) {
+  layers <- rlang::set_names(sass_bundle(x)$layers, NULL)
   sass_layer_struct(
-    defaults = join_non_null_values(layer2$defaults, layer1$defaults),
-    declarations = join_non_null_values(layer1$declarations, layer2$declarations),
-    rules = join_non_null_values(layer1$rules, layer2$rules),
-    html_deps = c(layer1$html_deps, layer2$html_deps),
-    file_attachments = join_attachments(layer1$file_attachments, layer2$file_attachments)
+    functions = pluck(layers, "functions"),
+    defaults = pluck(rev(layers), "defaults"),
+    mixins = pluck(layers, "mixins"),
+    rules = pluck(layers, "rules"),
+    declarations = pluck(layers, "declarations"),
+    html_deps = pluck(layers, "html_deps"),
+    file_attachments = pluck(layers, "file_attachments"),
+    validate = FALSE
   )
 }
-join_non_null_values <- function(x, y) {
-  ret <- dropNulls(list(x, y))
-  if (length(ret) == 0) return(NULL)
-  if (length(ret) == 1) return(ret[[1]])
-  ret
-}
-# attach2 takes precedence
-join_attachments <- function(attach1, attach2) {
-  # I thought about removing duplicates here, but it's hard to do so reliably
-  # because the paths can be files or directories.
-  c(attach1, attach2)
-}
 
-
-# Given the `input` to `sass()`, returns either NULL or a single sass_layer
-# that merges any sass_bundle found in the input
-# returns a single `sass_layer()` / `NULL`
-extract_layer <- function(input) {
-  if (is_sass_layer(input)) {
-    return(input)
-  }
-  if (is_sass_bundle(input)) {
-    return(as_sass_layer(input))
-  }
-  if (!identical(class(input), "list")) {
+pluck <- function(x, y) {
+  res <- dropNulls(lapply(x, `[[`, y))
+  if (length(res) == 0) {
     return(NULL)
   }
+  unlist(res, recursive = FALSE, use.names = TRUE)
+}
 
-  layers <- lapply(input, function(x) extract_layer(x))
-  layers <- dropNulls(layers)
-  if (length(layers) == 0) {
+extract_file_attachments <- function(x) {
+  if (is_sass_bundle_like(x)) {
+    return(as_sass_layer(x)$file_attachments)
+  }
+  if (!is.list(x)) {
     return(NULL)
   }
-  # convert to a sass layer object
-  as_sass_layer(
-    # merge all sass layers
-    sass_bundle(!!!layers)
+  unlist(
+    lapply(x, extract_file_attachments),
+    recursive = FALSE, use.names = TRUE
   )
 }
 
